#include <algorithm>
#include <cmath>
#include <cstddef>
#include <memory>
#include <arbor/mechanism_abi.h>
#include <arbor/math.hpp>

namespace arb {
namespace allen_catalogue {
namespace kernel_Nap {

using ::arb::math::exprelr;
using ::arb::math::safeinv;
using ::std::abs;
using ::std::cos;
using ::std::exp;
using ::std::max;
using ::std::min;
using ::std::pow;
using ::std::sin;
using ::std::sqrt;
using ::std::tanh;

static constexpr unsigned simd_width_ = 1;
static constexpr unsigned min_align_ = std::max(alignof(arb_value_type), alignof(arb_index_type));
using ::std::log;

#define PPACK_IFACE_BLOCK \
[[maybe_unused]] auto _pp_var_width                                                 = pp->width;\
[[maybe_unused]] auto _pp_var_n_detectors                                           = pp->n_detectors;\
[[maybe_unused]] auto _pp_var_dt                                                    = pp->dt;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_vec_ci                       = pp->vec_ci;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_v                        = pp->vec_v;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_i                        = pp->vec_i;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_g                        = pp->vec_g;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_temperature_degC             = pp->temperature_degC;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_diam_um                      = pp->diam_um;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_area_um2                     = pp->area_um2;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_time_since_spike             = pp->time_since_spike;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_node_index                   = pp->node_index;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_peer_index                   = pp->peer_index;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_multiplicity                 = pp->multiplicity;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_weight                       = pp->weight;\
[[maybe_unused]] auto& _pp_var_events                                               = pp->events;\
[[maybe_unused]] auto _pp_var_mechanism_id                                          = pp->mechanism_id;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_contiguous               = pp->index_constraints.n_contiguous;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_constant                 = pp->index_constraints.n_constant;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_independent              = pp->index_constraints.n_independent;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_none                     = pp->index_constraints.n_none;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_contiguous  = pp->index_constraints.contiguous;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_constant    = pp->index_constraints.constant;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_independent = pp->index_constraints.independent;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_none        = pp->index_constraints.none;\
[[maybe_unused]] auto const * const * _pp_var_random_numbers = pp->random_numbers;\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_h = pp->state_vars[0];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_qt = pp->state_vars[1];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_gbar = pp->parameters[0];\
[[maybe_unused]] auto& _pp_var_ion_na = pp->ion_states[0];\
[[maybe_unused]] auto* __restrict__ _pp_var_ion_na_index = pp->ion_states[0].index;\
//End of IFACEBLOCK


// interface methods
static void init(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        auto node_indexi_ = _pp_var_node_index[i_];
        arb_value_type v = _pp_var_vec_v[node_indexi_];
        arb_value_type celsius = _pp_var_temperature_degC[node_indexi_];
        _pp_var_qt[i_] = exp( 0.83290912293510388*((celsius- 21.0)* 0.10000000000000001));
        _pp_var_h[i_] =  1.0/( 1.0+exp((v+ 48.799999999999997)* 0.10000000000000001));
    }
    if (!_pp_var_multiplicity) return;
    for (arb_size_type ix = 0; ix < 1; ++ix) {
        for (arb_size_type iy = 0; iy < _pp_var_width; ++iy) {
            pp->state_vars[ix][iy] *= _pp_var_multiplicity[iy];
        }
    }
}

static void advance_state(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        auto node_indexi_ = _pp_var_node_index[i_];
        arb_value_type v = _pp_var_vec_v[node_indexi_];
        arb_value_type a_0_, hb, ha, ll0_, ba_0_, ll2_, ll4_, ll1_, ll3_, ll5_, ll6_;
        ll6_ =  0.;
        ll5_ =  0.;
        ll4_ =  0.;
        ll3_ =  0.;
        ll2_ =  0.;
        ll1_ =  0.;
        ll0_ =  0.;
        ll0_ = v+ 17.0;
        if (abs(ll0_* 0.21598272138228941)< 9.9999999999999995e-07) {
            ll1_ =  4.6299999999999999*( 1.0-ll0_* 0.21598272138228941* 0.5);
        }
        else {
            ll1_ = ll0_/(exp(ll0_* 0.21598272138228941)- 1.0);
        }
        ha =  2.88e-06*ll1_;
        ll2_ =  -(v+ 64.400000000000006);
        if (abs(ll2_* 0.38022813688212931)< 9.9999999999999995e-07) {
            ll3_ =  2.6299999999999999*( 1.0-ll2_* 0.38022813688212931* 0.5);
        }
        else {
            ll3_ = ll2_/(exp(ll2_* 0.38022813688212931)- 1.0);
        }
        hb =  6.9399999999999996e-06*ll3_;
        ll4_ =  1.0/( 1.0+exp((v+ 48.799999999999997)* 0.10000000000000001));
        a_0_ =  -1.0*(ha+hb)*_pp_var_qt[i_];
        ba_0_ = ll4_*(ha+hb)*_pp_var_qt[i_]/a_0_;
        ll5_ = a_0_*_pp_var_dt;
        ll6_ = ( 1.0+ 0.5*ll5_)/( 1.0- 0.5*ll5_);
        _pp_var_h[i_] =  -ba_0_+(_pp_var_h[i_]+ba_0_)*ll6_;
    }
}

static void compute_currents(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        auto ion_na_indexi_ = _pp_var_ion_na_index[i_];
        auto node_indexi_ = _pp_var_node_index[i_];
        arb_value_type current_ = 0;
        arb_value_type conductivity_ = 0;
        arb_value_type conductivity_na_ = 0;
        arb_value_type ena = _pp_var_ion_na.reversal_potential[ion_na_indexi_];
        arb_value_type ina = 0;
        arb_value_type v = _pp_var_vec_v[node_indexi_];
        arb_value_type m_inf, g;
        m_inf =  1.0/( 1.0+exp( -(v+ 52.600000000000001)* 0.21739130434782611));
        g = _pp_var_gbar[i_]*m_inf*_pp_var_h[i_];
        ina = g*(v-ena);
        conductivity_na_ = g;
        current_ = ina;
        conductivity_ = g;
        _pp_var_vec_i[node_indexi_] = fma(10.0*_pp_var_weight[i_], current_, _pp_var_vec_i[node_indexi_]);
        _pp_var_vec_g[node_indexi_] = fma(10.0*_pp_var_weight[i_], conductivity_, _pp_var_vec_g[node_indexi_]);
        _pp_var_ion_na.conductivity[ion_na_indexi_] = fma(10.0*_pp_var_weight[i_], conductivity_na_, _pp_var_ion_na.conductivity[ion_na_indexi_]);
        _pp_var_ion_na.current_density[ion_na_indexi_] = fma(10.0*_pp_var_weight[i_], ina, _pp_var_ion_na.current_density[ion_na_indexi_]);
    }
}

static void write_ions(arb_mechanism_ppack* pp) {
}

static void apply_events(arb_mechanism_ppack*, arb_deliverable_event_stream*) {}

static void post_event(arb_mechanism_ppack*) {}
#undef PPACK_IFACE_BLOCK
} // namespace kernel_Nap
} // namespace allen_catalogue
} // namespace arb

extern "C" {
  arb_mechanism_interface* make_arb_allen_catalogue_Nap_interface_multicore() {
    static arb_mechanism_interface result;
    result.partition_width = arb::allen_catalogue::kernel_Nap::simd_width_;
    result.backend = arb_backend_kind_cpu;
    result.alignment = arb::allen_catalogue::kernel_Nap::min_align_;
    result.init_mechanism = arb::allen_catalogue::kernel_Nap::init;
    result.compute_currents = arb::allen_catalogue::kernel_Nap::compute_currents;
    result.apply_events = arb::allen_catalogue::kernel_Nap::apply_events;
    result.advance_state = arb::allen_catalogue::kernel_Nap::advance_state;
    result.write_ions = arb::allen_catalogue::kernel_Nap::write_ions;
    result.post_event = arb::allen_catalogue::kernel_Nap::post_event;
    return &result;
  }
}

